/* -*-c-*-
 *
 * $Id: fw.c,v 1.1 1999/07/01 08:56:23 mdw Exp $
 *
 * Port forwarding thingy
 *
 * (c) 1999 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of the `fw' port forwarder.
 *
 * `fw' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * `fw' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with `fw'; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: fw.c,v $
 * Revision 1.1  1999/07/01 08:56:23  mdw
 * Initial revision
 *
 */

/*----- Header files ------------------------------------------------------*/

#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <syslog.h>

#include <mLib/mdwopt.h>
#include <mLib/quis.h>
#include <mLib/report.h>
#include <mLib/sel.h>
#include <mLib/sub.h>

#include "acl.h"
#include "bres.h"
#include "conf.h"
#include "listener.h"
#include "scan.h"

/*----- Static variables --------------------------------------------------*/

sel_state *sel;				/* Multiplexor for nonblocking I/O */

/*----- Main code ---------------------------------------------------------*/

/* --- Standard GNU help options --- */

static void version(FILE *fp)
{
  fprintf(fp, "%s version " VERSION "\n", QUIS);
}

static void usage(FILE *fp)
{
  fprintf(stderr, "Usage: %s [-d] [-f file] [config statements...]\n",
	  QUIS);
}

static void help(FILE *fp)
{
  version(fp);
  fputc('\n', fp);
  usage(fp);
  fputs("\n\
A fairly full-featured port-forwarder.  Options available are:\n\
\n\
-h, --help		Display this help message.\n\
-v, --version		Display the program's version number.\n\
-u, --usage		Display a terse usage summary.\n\
\n\
-f, --file=FILE		Read configuration from a file.\n\
-d, --dump		Dump the configuration to standard output.\n\
\n\
Configuration may be supplied in one or more configuration files, or on\n\
the command line (or both).  If no `-f' option is present, and no\n\
configuration is given on the command line, the standard input stream is\n\
read.\n\
\n\
Configuration is free-form.  Comments begin with a `#' character and\n\
continue to the end of the line.  When reading from the command line,\n\
each argument is considered a separate line.  See the manual page for the\n\
complete syntax.\n\
", fp);
}

/* --- @main@ --- *
 *
 * Arguments:	@int argc@ = number of command line arguments
 *		@char *argv[]@ = vector of argument strings
 *
 * Returns:	---
 *
 * Use:		Simple port-forwarding server.
 */

int main(int argc, char *argv[])
{
  unsigned f = 0;
  sel_state sst;

  enum {
    f_bogus = 1,
    f_file = 2,
    f_dump = 4,
    f_fork = 8
  };

  /* --- Initialize things --- */

  ego(argv[0]);
  sel = &sst;
  sel_init(sel);
  sub_init();
  bres_init(sel);

  /* --- Parse command line options --- */

  for (;;) {
    static struct option opts[] = {

      /* --- Standard GNU help options --- */

      { "help", 	0,		0,	'h' },
      { "version",	0,		0,	'v' },
      { "usage",	0,		0,	'u' },

      /* --- Other useful arguments --- */

      { "file",		OPTF_ARGREQ,	0,	'f' },
      { "dump",		0,		0,	'd' },
      { "fork",		0,		0,	'b' },
      { "background",	0,		0,	'b' },

      /* --- Magic terminator --- */

      { 0,		0,		0,	0 }
    };
    int i = mdwopt(argc, argv, "hvu f:db", opts, 0, 0, 0);

    if (i < 0)
      break;
    switch (i) {
      case 'h':
	help(stdout);
	exit(0);
	break;
      case 'v':
	version(stdout);
	exit(0);
	break;
      case 'u':
	usage(stdout);
	exit(0);
	break;
      case 'f': {
	scan_filectx ctx;
	FILE *fp;
	if ((fp = fopen(optarg, "r")) == 0)
	  die(1, "couldn't open file `%s': %s", optarg, strerror(errno));
	scan_fileinit(&ctx, fp, optarg);
	conf_parse(&ctx);
	fclose(fp);
	f |= f_file;
      } break;
      case 'd':
	f |= f_dump;
	break;
      case 'b':
	f |= f_fork;
	break;
      default:
	f |= f_bogus;
	break;
    }
  }

  if (f & f_bogus) {
    usage(stderr);
    exit(1);
  }

  /* --- Deal with the remaining arguments --- */

  if (optind == argc) {
    if (f & f_file)
      /* Cool */;
    else if (isatty(STDIN_FILENO)) {
      moan("no configuration given and stdin is a terminal.");
      moan("type `%s --help' for usage information.", QUIS);
      exit(1);
    } else {
      scan_filectx ctx;
      scan_fileinit(&ctx, stdin, "<stdin>");
      conf_parse(&ctx);
    }
  } else {
    scan_argvctx ctx;
    scan_argvinit(&ctx, argv + optind);
    conf_parse(&ctx);
  }

  /* --- Dump out the state --- */

  if (f & f_dump) {
    sel_file *s;
    fputs("global acl:\n", stdout);
    acl_dump(0, stdout);
    for (s = sel->files; s; s = s->next)
      listener_dump((listener *)s, stdout);
    exit(0);
  }

  /* --- Fork into the background --- */

  if (f & f_fork) {
    pid_t kid;

    kid = fork();
    if (kid == -1)
      die(1, "couldn't fork: %s", strerror(errno));
    if (kid != 0)
      _exit(0);

    close(0); close(1); close(2);
    setsid();

    kid = fork();
    if (kid != 0)
      _exit(0);
  }

  /* --- Let rip --- */

  openlog(QUIS, 0, LOG_DAEMON);
  for (;;)
    sel_select(sel);
  return (0);
}

/*----- That's all, folks -------------------------------------------------*/
